package rabinizer.exec;

import rabinizer.formulas.*;

import java.util.*;

import net.sf.javabdd.*;

public class GeneralizedRabinAcceptanceCondition {

    private List<Formula> aRFsAll = null;
    private List<Formula> aRFsPowersetElement = null;
    private List<Formula> promises = null;
    private List<Formula> negativePromises = null;
    private Map<Formula, Formula> aRFsPowersetElementUnfold = null;
    private Map<Formula, Integer> pi = null;
    private KDFA kdfaA = null;
    public Set<Tuple<Map<Tuple<Tuple<Formula, KState>, Tuple<Formula, KState>>, ValuationSet>, List<Map<Tuple<Tuple<Formula, KState>, Tuple<Formula, KState>>, ValuationSet>>>> condition = null;
    public Map<Map<Tuple<Tuple<Formula, KState>, Tuple<Formula, KState>>, ValuationSet>, Integer> edgesToInt = null;
    private Integer format;
    private Set<Formula> xgEntails = null;

    GeneralizedRabinAcceptanceCondition(List<Formula> aRFs, KDFA _kdfaA, Integer form) {
        setFormat(form);
        kdfaA = _kdfaA;
        aRFsAll = aRFs;
        aRFsPowersetElementUnfold = new HashMap<Formula, Formula>();
        condition = new HashSet<Tuple<Map<Tuple<Tuple<Formula, KState>, Tuple<Formula, KState>>, ValuationSet>, List<Map<Tuple<Tuple<Formula, KState>, Tuple<Formula, KState>>, ValuationSet>>>>();
        xgEntails = new HashSet<Formula>();

        Misc.verboseln("\nGeneralized Rabin Acceptance Condition for " + _kdfaA.phi);
        iteratePowersetAlwaysRFs(aRFs, 0, aRFs.size());

        Misc.verboseln("\nRemoving redundancy from Gen Rabin Acceptance Condition for " + _kdfaA.phi);
        removeRedundancy();
    }

    //Iterate over all subsets of set containing all G subformulas
    private void iteratePowersetAlwaysRFs(List<Formula> powersetElement, int pos, int count) {
        if (count <= 0) {
            aRFsPowersetElement = powersetElement;
            aRFsPowersetElementUnfold.clear();
            for (Formula f : powersetElement) {
                aRFsPowersetElementUnfold.put(f.unfoldNoG(), f);
            }
            pi = new HashMap<Formula, Integer>(aRFsPowersetElement.size());
            promises = new ArrayList<Formula>(aRFsPowersetElement.size());
            negativePromises = new ArrayList<Formula>(aRFsAll.size() - aRFsPowersetElement.size());
            Misc.verbose("\nPromise set: ");

            // massage the promises
            // always- promises are applications of G , construct them from aRFs (add G operator ahead the formula)
            for (int i = 0; i < aRFsPowersetElement.size(); i++) {
                Formula p = new GOperator(aRFsPowersetElement.get(i));
                Misc.verbose(" " + p + ",");
                promises.add(p);
            }
            Misc.verboseln("\n************");
            
            Misc.verboseln("\nNegative promises:");
            for (int i = 0; i < aRFsAll.size(); i++) {
                Formula p = new GOperator(aRFsAll.get(i));
                if(!promises.contains(p)){
                    Misc.verbose(" " + p + ",");
                    negativePromises.add(p);
                }
            }

            xgEntails.clear();
            Misc.verboseln("\n\nSegment entailments:\n");
            for (SegmentAutomaton sa : kdfaA.segmentAutomata) {
                for (Formula f : sa.states) {
                    if (entails(null, promises, null, f)) {
                        xgEntails.add(f);
                    }
                }
            }

            Misc.verboseln("");

            iterateOverRanks(0);
        } else {
            // keep element in position pos
            iteratePowersetAlwaysRFs(powersetElement, pos + 1, count - 1);
            // remove element in position pos 
            List<Formula> currPosRemoved = new ArrayList<Formula>(powersetElement.size());
            currPosRemoved.addAll(powersetElement);
            currPosRemoved.remove(pos);
            iteratePowersetAlwaysRFs(currPosRemoved, pos, count - 1);
        }
    }

    //iterate over all mappings pi: aRFsPowersetElement -> Integers, so that the value is less than max rank for the formula 
    private void iterateOverRanks(int count) {
        if (count >= aRFsPowersetElement.size()) {
            rabinPair();
        } else {
            // iterate over all possible ranks of automata over any word
            int max = kdfaA.sAToMaxSinkRank.get(aRFsPowersetElement.get(count).unfoldNoG());
            Misc.verboseln("Segment " + count + " with max rank: " + max);
            for (int i = 1; i <= max; i++) {
                Misc.verboseln("  Rank " + i);
                pi.put(aRFsPowersetElement.get(count), i);
                iterateOverRanks(count + 1);
            }
        }
    }

    //int counter = 0;
    private void rabinPair() {
        
        //Misc.verbose("ENTAILS: "+xgEntails);

        // Set avoid
        Misc.verboseln("Set avoid:");

        Map<Tuple<Tuple<Formula, KState>, Tuple<Formula, KState>>, ValuationSet> avoid = new HashMap<Tuple<Tuple<Formula, KState>, Tuple<Formula, KState>>, ValuationSet>();

        for (Map.Entry<Tuple<Tuple<Formula, KState>, Tuple<Formula, KState>>, ValuationSet> edge : kdfaA.edges.entrySet()) {
            Tuple<Tuple<Formula, KState>, Tuple<Formula, KState>> edgeKey = edge.getKey();
            Tuple<Formula, KState> state2 = edgeKey.right;
            List<Formula> relevantSegmentStates = new ArrayList<Formula>();
            for (Formula promise : aRFsPowersetElement) {
                Map<Formula, SegmentAutomaton> sFToSA = (state2.right).segmentFormulasToSA;
                if (sFToSA.keySet().contains(promise.unfoldNoG())) {
                    for (Formula f : state2.right.state.get(sFToSA.get(promise.unfoldNoG())).keySet()) {
                        if (state2.right.state.get(sFToSA.get(promise.unfoldNoG())).get(f) >= pi.get(promise)) {
                            relevantSegmentStates.add(f);
                            relevantSegmentStates.add(f.unfold());
                        }
                    }
                }
            }
            if (!entails(relevantSegmentStates, promises, negativePromises, state2.left)) {
                //Misc.verboseln("  new edge");
                avoid.put(edge.getKey(), edge.getValue());
            }
        }
        Misc.verboseln("avoid:" + avoid);

        //Set fail
        Misc.verboseln("Set fail:");
        Map<Tuple<Tuple<Formula, KState>, Tuple<Formula, KState>>, ValuationSet> fail = new HashMap<Tuple<Tuple<Formula, KState>, Tuple<Formula, KState>>, ValuationSet>();

        for (Map.Entry<Tuple<Tuple<Formula, KState>, Tuple<Formula, KState>>, ValuationSet> edge : kdfaA.edges.entrySet()) {
            Tuple<Tuple<Formula, KState>, Tuple<Formula, KState>> edgeKey = edge.getKey();
            Tuple<Formula, KState> state1 = edgeKey.left;

            Map<Formula, SegmentAutomaton> sFToSA = (state1.right).segmentFormulasToSA;

            breakloop:
            {
                for (Formula rel : sFToSA.keySet()) {
                    if (aRFsPowersetElementUnfold.keySet().contains(rel)) {
                        //Misc.verboseln("  relevant slave for "+rel);
                        SegmentAutomaton sa = sFToSA.get(rel);
                        for (Formula q : state1.right.state.get(sa).keySet()) {
                            for (Valuation v : edge.getValue()) {
                                if (sa.finalStates.contains(sa.tranFunction.get(q).get(v)) && (!xgEntails.contains(sa.tranFunction.get(q).get(v)))) {
                                    //Misc.verboseln("  new accepting edge with target " + sa.tranFunction.get(q).get(v));
                                    if (fail.keySet().contains(edgeKey)) {
                                        ValuationSet label = fail.get(edgeKey);
                                        label.add(v);
                                        fail.put(edgeKey, label);
                                    } else {
                                        ValuationSet label = new ValuationSet();
                                        label.add(v);
                                        fail.put(edgeKey, label);
                                    }
                                } else
									;//Misc.verboseln("  not interesting edge with target " + sa.tranFunction.get(q).get(v));
                            }
                        }
                    }
                }
            }
        }

        Misc.verboseln("fail:" + fail);

        // Set buy(pi)
        Misc.verboseln("Set buy(pi):");
        Map<Tuple<Tuple<Formula, KState>, Tuple<Formula, KState>>, ValuationSet> buyPi = new HashMap<Tuple<Tuple<Formula, KState>, Tuple<Formula, KState>>, ValuationSet>();

        for (Map.Entry<Tuple<Tuple<Formula, KState>, Tuple<Formula, KState>>, ValuationSet> edge : kdfaA.edges.entrySet()) {
            Tuple<Tuple<Formula, KState>, Tuple<Formula, KState>> edgeKey = edge.getKey();
            Tuple<Formula, KState> state1 = edgeKey.left;

            Map<Formula, SegmentAutomaton> sFToSA = (state1.right).segmentFormulasToSA;

            breakloop:
            {
                for (Formula rel : sFToSA.keySet()) {
                    if (aRFsPowersetElementUnfold.keySet().contains(rel)) {
                        SegmentAutomaton sa = sFToSA.get(rel);
                        for (Formula q : state1.right.state.get(sa).keySet()) {
                            if (state1.right.state.get(sa).get(q) < pi.get(aRFsPowersetElementUnfold.get(rel))) {
                                for (Formula q2 : state1.right.state.get(sa).keySet()) {
                                    if (!q.equals(q2)) {
                                        for (Valuation v : edge.getValue()) {
                                            //if(sa.tranFunction.get(q).get(v).equals(sa.tranFunction.get(q2).get(v))&& (!entails(null, promises, sa.tranFunction.get(q).get(v)))) {
                                            if (sa.tranFunction.get(q).get(v).equals(sa.tranFunction.get(q2).get(v)) && (!xgEntails.contains(sa.tranFunction.get(q).get(v)))) {
                                                //Misc.verboseln("  new buy edge" + q + "/" + q2 + " under " + v);
                                                if (buyPi.keySet().contains(edgeKey)) {
                                                    ValuationSet label = buyPi.get(edgeKey);
                                                    label.add(v);
                                                    buyPi.put(edgeKey, label);
                                                } else {
                                                    ValuationSet label = new ValuationSet();
                                                    label.add(v);
                                                    buyPi.put(edgeKey, label);
                                                }
                                            }
                                        }
                                    } else {
                                        for (Valuation v : edge.getValue()) {
                                            //if(sa.tranFunction.get(q).get(v).equals(sa.tranFunction.get(q2).get(v))&& (!entails(null, promises, sa.tranFunction.get(q).get(v)))) {
                                            if (sa.tranFunction.get(q).get(v).equals(sa.initialState) && (!xgEntails.contains(sa.tranFunction.get(q).get(v)))) {
                                                //Misc.verboseln("  new buy edge" + q + "/" + q2 + " under " + v);
                                                if (buyPi.keySet().contains(edgeKey)) {
                                                    ValuationSet label = buyPi.get(edgeKey);
                                                    label.add(v);
                                                    buyPi.put(edgeKey, label);
                                                } else {
                                                    ValuationSet label = new ValuationSet();
                                                    label.add(v);
                                                    buyPi.put(edgeKey, label);
                                                }
                                            }
                                        }
                                    }

                                }

                            }
                        }
                    }
                }
            }
        }

        Misc.verboseln("buy(pi):" + buyPi);

        // Set succeed(pi)
        Misc.verboseln("Set succeed(pi):");
        List<Map<Tuple<Tuple<Formula, KState>, Tuple<Formula, KState>>, ValuationSet>> succeedList = new ArrayList<Map<Tuple<Tuple<Formula, KState>, Tuple<Formula, KState>>, ValuationSet>>();

        for (Formula rel : aRFsPowersetElement) {
            Misc.verboseln("  " + rel);
            Map<Tuple<Tuple<Formula, KState>, Tuple<Formula, KState>>, ValuationSet> succeed = new HashMap<Tuple<Tuple<Formula, KState>, Tuple<Formula, KState>>, ValuationSet>();
            for (Map.Entry<Tuple<Tuple<Formula, KState>, Tuple<Formula, KState>>, ValuationSet> edge : kdfaA.edges.entrySet()) {
                Tuple<Tuple<Formula, KState>, Tuple<Formula, KState>> edgeKey = edge.getKey();
                Tuple<Formula, KState> state1 = edgeKey.left;

                Map<Formula, SegmentAutomaton> sFToSA = (state1.right).segmentFormulasToSA;
                SegmentAutomaton sa = sFToSA.get(rel.unfoldNoG());

                breakloop:
                {
                    if (sFToSA.keySet().contains(rel.unfoldNoG())) {
                        for (Formula q : state1.right.state.get(sa).keySet()) {
                            if (state1.right.state.get(sa).get(q) == pi.get(rel)) {
                                for (Valuation v : edge.getValue()) {
                                    //if(sa.finalStates.contains(sa.tranFunction.get(q).get(v)) && (entails(null, promises, sa.tranFunction.get(q).get(v)))&& (!entails(null, promises, q))) {
                                    if (xgEntails.contains(sa.initialState) || (xgEntails.contains(sa.tranFunction.get(q).get(v)) && !xgEntails.contains(q))) {
                                        //Misc.verboseln("  new accepting edge with target " + sa.tranFunction.get(q).get(v));
                                        //Misc.verboseln("       " + edgeKey.left + " " + v);
                                        if (succeed.keySet().contains(edgeKey)) {
                                            ValuationSet label = succeed.get(edgeKey);
                                            label.add(v);
                                            succeed.put(edgeKey, label);
                                        } else {
                                            ValuationSet label = new ValuationSet();
                                            label.add(v);
                                            succeed.put(edgeKey, label);
                                        }
                                    } else ;// Misc.verboseln("  not interesting edge with target " + sa.tranFunction.get(q).get(v));
                                }
                            }
                        }
                    }
                }
            }
            if (!xgEntails.contains(rel.unfoldNoG())) //TODO fixes e.g. FGGa //  rel.isProgressFormula() ||
            {
                succeedList.add(succeed);
            }
        }

        Misc.verboseln("succeed(pi)list:" + succeedList);

        // add fail to avoid
        for (Map.Entry<Tuple<Tuple<Formula, KState>, Tuple<Formula, KState>>, ValuationSet> edge : fail.entrySet()) {
            Tuple<Tuple<Formula, KState>, Tuple<Formula, KState>> edgeKey = edge.getKey();

            if (avoid.keySet().contains(edgeKey)) {
                ValuationSet label = avoid.get(edgeKey);
                label.addAll(edge.getValue());
                avoid.put(edgeKey, label);
            } else {
                ValuationSet label = new ValuationSet();
                label.addAll(edge.getValue());
                avoid.put(edgeKey, label);
            }
        }

        // add buy to avoid
        for (Map.Entry<Tuple<Tuple<Formula, KState>, Tuple<Formula, KState>>, ValuationSet> edge : buyPi.entrySet()) {
            Tuple<Tuple<Formula, KState>, Tuple<Formula, KState>> edgeKey = edge.getKey();

            if (avoid.keySet().contains(edgeKey)) {
                ValuationSet label = avoid.get(edgeKey);
                label.addAll(edge.getValue());
                avoid.put(edgeKey, label);
            } else {
                ValuationSet label = new ValuationSet();
                label.addAll(edge.getValue());
                avoid.put(edgeKey, label);
            }
        }

        // the hack - for empty succeedList
        if (succeedList.isEmpty()) {
            List<Map<Tuple<Tuple<Formula, KState>, Tuple<Formula, KState>>, ValuationSet>> allEdges = new ArrayList<Map<Tuple<Tuple<Formula, KState>, Tuple<Formula, KState>>, ValuationSet>>();
            allEdges.add(kdfaA.edges);
            succeedList.addAll(allEdges);
        };
        Tuple<Map<Tuple<Tuple<Formula, KState>, Tuple<Formula, KState>>, ValuationSet>, List<Map<Tuple<Tuple<Formula, KState>, Tuple<Formula, KState>>, ValuationSet>>> pair
            = new Tuple<Map<Tuple<Tuple<Formula, KState>, Tuple<Formula, KState>>, ValuationSet>, List<Map<Tuple<Tuple<Formula, KState>, Tuple<Formula, KState>>, ValuationSet>>>(avoid, succeedList);
        Misc.verboseln(pairToString(pair));
        condition.add(pair);
    }

    public static boolean entails(List<Formula> regularPremises, List<Formula> promisePremises, List<Formula> negativePromisePremises, Formula consequence) {
        Formula nakedConsequence = consequence.removeXsFromCurrentBooleanAtoms();  // if you know what I mean :3
        Formula premise = Misc.fTrue;
        if (regularPremises != null) {
            for (Formula rp : regularPremises) {
                premise = new Conjunction(premise, rp.removeXsFromCurrentBooleanAtoms());  // TODO check the X-removal
            }
        }
        if (promisePremises != null) {
            for (Formula pp : promisePremises) {
                premise = new Conjunction(premise, pp);
            }
        }
        if (negativePromisePremises != null) {
            for (Formula np : negativePromisePremises) {
                premise = new Conjunction(premise, new FormalNegation(np));
            }
        }
        //		Misc.verboseln("\n\n");

        Misc.verboseln("    premise          : " + premise);
        Misc.verboseln("    consequence      : " + consequence);
        Misc.verboseln("    nakedConsequence : " + nakedConsequence);

        //		Misc.verboseln("premise repr: "+premise.representative());
        //		Misc.verboseln("nakedC  repr: "+nakedConsequence.representative());
        //
        //		Misc.verboseln("premise bdd: "+premise.bdd());
        //		Misc.verboseln("nakedC  bdd: "+nakedConsequence.bdd());
        //		Misc.verboseln("\n\n");
        BDD b = premise.bdd().imp(nakedConsequence.bdd());
        boolean entails = b.equals(Misc.bddFactory.one());
        Misc.verboseln("    entails          : " + entails);
        return entails;
    }

    public int size() {
        return condition.size();
    }

    public String toString() {
        return toString(condition);
    }

    public String toString(Set<Tuple<Map<Tuple<Tuple<Formula, KState>, Tuple<Formula, KState>>, ValuationSet>, List<Map<Tuple<Tuple<Formula, KState>, Tuple<Formula, KState>>, ValuationSet>>>> theCondition) {
        int rpc = 0;
        String r = "Generalized Rabin pair count: " + theCondition.size() + "\n";

        for (Tuple<Map<Tuple<Tuple<Formula, KState>, Tuple<Formula, KState>>, ValuationSet>, List<Map<Tuple<Tuple<Formula, KState>, Tuple<Formula, KState>>, ValuationSet>>> pair : theCondition) {
            r += "Pair " + (++rpc) + ":\n";
            if (format == 1) { // format is dotty

                r += pairToStringNumbered(pair);
            } else if (format == 3) {
                r += pairToStringNotBDD(pair);
            } else {
                r += pairToString(pair); //format is dotty-full
            }
        }
        return r;
    }

    public String toNumericalString() {
        String r = "";
        String id;
        edgesToInt = new HashMap<Map<Tuple<Tuple<Formula, KState>, Tuple<Formula, KState>>, ValuationSet>, Integer>();
        HashMap<Tuple<Tuple<Formula, KState>, Tuple<Formula, KState>>, ValuationSet> emptySet = new HashMap<Tuple<Tuple<Formula, KState>, Tuple<Formula, KState>>, ValuationSet>();
        edgesToInt.put(emptySet, 0);
        edgesToInt.put(kdfaA.edges, -1);
        Integer rpc = 1;
        Integer rpc2 = 0;

        for (Tuple<Map<Tuple<Tuple<Formula, KState>, Tuple<Formula, KState>>, ValuationSet>, List<Map<Tuple<Tuple<Formula, KState>, Tuple<Formula, KState>>, ValuationSet>>> pair : condition) {
            if (rpc2 > 0) {
                r += " | ";
            }
            rpc2++;
            if (! edgesToInt.keySet().contains(pair.left)) {
                edgesToInt.put(pair.left, rpc);
                rpc++;
            }
            id = edgesToInt.get(pair.left).toString();
            if(id.startsWith("-1")) id = "!0";
            r += "(Fin(" + id + ")";

            for (Map<Tuple<Tuple<Formula, KState>, Tuple<Formula, KState>>, ValuationSet> reachSet : pair.right) {
                if (! edgesToInt.keySet().contains(reachSet)) {
                    edgesToInt.put(reachSet, rpc);
                    rpc++;
                }
                id = edgesToInt.get(reachSet).toString();
                if(id.startsWith("-1")) id = "!0";
                r += "&Inf(" + id + ")";
            }
            r += ")";
        }
        edgesToInt.remove(kdfaA.edges);
        return rpc + " " + r;
    }

    private String pairToString(Tuple<Map<Tuple<Tuple<Formula, KState>, Tuple<Formula, KState>>, ValuationSet>, List<Map<Tuple<Tuple<Formula, KState>, Tuple<Formula, KState>>, ValuationSet>>> pair) {
        String r = "[\n";
        for (Map.Entry<Tuple<Tuple<Formula, KState>, Tuple<Formula, KState>>, ValuationSet> edge : pair.left.entrySet()) {
            r += " " + edge.getKey().left + ": ";
            //r += " "+edge.getValue()+" ";  //\n
            r += BDDService.BDDtoString(BDDService.valuationSetToBDD(edge.getValue())) + "\n";
        }

        r += "]\n[\n";
        for (Map<Tuple<Tuple<Formula, KState>, Tuple<Formula, KState>>, ValuationSet> reachSet : pair.right) {
            r += " [\n";
            for (Map.Entry<Tuple<Tuple<Formula, KState>, Tuple<Formula, KState>>, ValuationSet> edge : reachSet.entrySet()) {
                r += " " + edge.getKey().left + ": ";
                //r += " "+edge.getValue()+"\n";
                r += BDDService.BDDtoString(BDDService.valuationSetToBDD(edge.getValue())) + "\n";
            }
            r += " ]\n";
        }
        r += "]\n";
        return r;
    }

    private String pairToStringNumbered(Tuple<Map<Tuple<Tuple<Formula, KState>, Tuple<Formula, KState>>, ValuationSet>, List<Map<Tuple<Tuple<Formula, KState>, Tuple<Formula, KState>>, ValuationSet>>> pair) {
        String r = "[\n";
        Map<Tuple<Formula, KState>, ValuationSet> simplePairs = new HashMap<Tuple<Formula, KState>, ValuationSet>();
        for (Map.Entry<Tuple<Tuple<Formula, KState>, Tuple<Formula, KState>>, ValuationSet> edge : pair.left.entrySet()) {
            ValuationSet valSet = new ValuationSet();
            if (simplePairs.containsKey(edge.getKey().left)) {

                for (Valuation v : simplePairs.get(edge.getKey().left)) {
                    valSet.add(v);

                }
                valSet.addAll(edge.getValue());
                simplePairs.put(edge.getKey().left, valSet);
                //Misc.verboseln(simplePairs.get(edge.getKey()));
            } else {
                simplePairs.put(edge.getKey().left, edge.getValue());
            }
        }
        for (Map.Entry<Tuple<Formula, KState>, ValuationSet> edge : simplePairs.entrySet()) {
            r += " " + kdfaA.statesToNumbers.get(edge.getKey()) + ": ";
            //r += " "+edge.getValue();
            r += BDDService.BDDtoString(BDDService.valuationSetToBDD(edge.getValue())) + "\n";//TODO output bdd representation
        }
        r += "]\n[\n";
        Map<Tuple<Formula, KState>, ValuationSet> simplePairsRight = new HashMap<Tuple<Formula, KState>, ValuationSet>();
        for (Map<Tuple<Tuple<Formula, KState>, Tuple<Formula, KState>>, ValuationSet> reachSet : pair.right) {
            r += " [\n";
            for (Map.Entry<Tuple<Tuple<Formula, KState>, Tuple<Formula, KState>>, ValuationSet> edge : reachSet.entrySet()) {
                ValuationSet valSet = new ValuationSet();
                if (simplePairs.containsKey(edge.getKey().left)) {
                    for (Valuation v : simplePairs.get(edge.getKey().left)) {
                        valSet.add(v);

                    }
                    valSet.addAll(edge.getValue());
                    simplePairsRight.put(edge.getKey().left, valSet);
                    //Misc.verboseln(simplePairs.get(edge.getKey()));
                } else {
                    simplePairsRight.put(edge.getKey().left, edge.getValue());
                }
            }
            for (Map.Entry<Tuple<Formula, KState>, ValuationSet> edge : simplePairsRight.entrySet()) {
                r += " " + kdfaA.statesToNumbers.get(edge.getKey()) + ": ";
                //r += " "+edge.getValue()+"\n";
                r += BDDService.BDDtoString(BDDService.valuationSetToBDD(edge.getValue())) + "\n";
            }
            r += " ]\n";
        }
        r += "]\n";
        return r;
    }

    private String pairToStringNotBDD(Tuple<Map<Tuple<Tuple<Formula, KState>, Tuple<Formula, KState>>, ValuationSet>, List<Map<Tuple<Tuple<Formula, KState>, Tuple<Formula, KState>>, ValuationSet>>> pair) {
        String r = "[\n";
        Map<Tuple<Formula, KState>, ValuationSet> simplePairs = new HashMap<Tuple<Formula, KState>, ValuationSet>();
        for (Map.Entry<Tuple<Tuple<Formula, KState>, Tuple<Formula, KState>>, ValuationSet> edge : pair.left.entrySet()) {
            ValuationSet valSet = new ValuationSet();
            if (simplePairs.containsKey(edge.getKey().left)) {

                for (Valuation v : simplePairs.get(edge.getKey().left)) {
                    valSet.add(v);

                }
                valSet.addAll(edge.getValue());
                simplePairs.put(edge.getKey().left, valSet);
                //Misc.verboseln(simplePairs.get(edge.getKey()));
            } else {
                simplePairs.put(edge.getKey().left, edge.getValue());
            }
        }
        for (Map.Entry<Tuple<Formula, KState>, ValuationSet> edge : simplePairs.entrySet()) {
            r += " " + kdfaA.statesToNumbers.get(edge.getKey()) + ": ";
            //r += " "+edge.getValue();
            r += edge.getValue() + "\n";//TODO output bdd representation
        }
        r += "]\n[\n";
        Map<Tuple<Formula, KState>, ValuationSet> simplePairsRight = new HashMap<Tuple<Formula, KState>, ValuationSet>();
        for (Map<Tuple<Tuple<Formula, KState>, Tuple<Formula, KState>>, ValuationSet> reachSet : pair.right) {
            r += " [\n";
            for (Map.Entry<Tuple<Tuple<Formula, KState>, Tuple<Formula, KState>>, ValuationSet> edge : reachSet.entrySet()) {
                ValuationSet valSet = new ValuationSet();
                if (simplePairs.containsKey(edge.getKey().left)) {
                    for (Valuation v : simplePairs.get(edge.getKey().left)) {
                        valSet.add(v);

                    }
                    valSet.addAll(edge.getValue());
                    simplePairsRight.put(edge.getKey().left, valSet);
                    //Misc.verboseln(simplePairs.get(edge.getKey()));
                } else {
                    simplePairsRight.put(edge.getKey().left, edge.getValue());
                }
            }
            for (Map.Entry<Tuple<Formula, KState>, ValuationSet> edge : simplePairsRight.entrySet()) {
                r += " " + kdfaA.statesToNumbers.get(edge.getKey()) + ": ";
                //r += " "+edge.getValue()+"\n";
                r += edge.getValue() + "\n";
            }
            r += " ]\n";
        }
        r += "]\n";
        return r;
    }

    private void removeRedundancy() {
        Misc.verboseln("0. Raw Generalized Rabin Acceptance Condition\n");
        Misc.verboseln(toString(condition));

        Misc.verboseln("1. Removing avoid states from reach sets by mapping each gen pair (F, {i1,...,in}) |-> (F, {i1\\F,...,in\\F})\n");
        Set<Tuple<Map<Tuple<Tuple<Formula, KState>, Tuple<Formula, KState>>, ValuationSet>, List<Map<Tuple<Tuple<Formula, KState>, Tuple<Formula, KState>>, ValuationSet>>>> step1 = new HashSet<Tuple<Map<Tuple<Tuple<Formula, KState>, Tuple<Formula, KState>>, ValuationSet>, List<Map<Tuple<Tuple<Formula, KState>, Tuple<Formula, KState>>, ValuationSet>>>>();
        for (Tuple<Map<Tuple<Tuple<Formula, KState>, Tuple<Formula, KState>>, ValuationSet>, List<Map<Tuple<Tuple<Formula, KState>, Tuple<Formula, KState>>, ValuationSet>>> pair : condition) {
            List<Map<Tuple<Tuple<Formula, KState>, Tuple<Formula, KState>>, ValuationSet>> newSucceedList = new ArrayList<Map<Tuple<Tuple<Formula, KState>, Tuple<Formula, KState>>, ValuationSet>>();
            for (Map<Tuple<Tuple<Formula, KState>, Tuple<Formula, KState>>, ValuationSet> i : pair.right) {
                Map<Tuple<Tuple<Formula, KState>, Tuple<Formula, KState>>, ValuationSet> newSucceedListERF = new HashMap<Tuple<Tuple<Formula, KState>, Tuple<Formula, KState>>, ValuationSet>();
                //newSucceedListERF.addAll(i);
                for (Map.Entry<Tuple<Tuple<Formula, KState>, Tuple<Formula, KState>>, ValuationSet> edge : i.entrySet()) {
                    Tuple<Tuple<Formula, KState>, Tuple<Formula, KState>> edgeKey = edge.getKey();
                    /*
                     if(newSucceedListERF.keySet().contains(edgeKey)){
                     ValuationSet label = newSucceedListERF.get(edgeKey);
                     label.addAll(edge.getValue());
                     newSucceedListERF.put(edgeKey, label);
                     } else {
                     ValuationSet label = new ValuationSet();
                     label.addAll(edge.getValue());
                     newSucceedListERF.put(edgeKey, label);
                     }
                     */
                    ValuationSet label = new ValuationSet();
                    label.addAll(edge.getValue());
                    if (pair.left.keySet().contains(edgeKey)) {
                        label.removeAll(pair.left.get(edgeKey));
                    }
                    if (!label.isEmpty()) {
                        newSucceedListERF.put(edgeKey, label);
                    }
                }
                // if (!newReachEventuallyERF.isEmpty())
                newSucceedList.add(newSucceedListERF);
            }
            step1.add(new Tuple<Map<Tuple<Tuple<Formula, KState>, Tuple<Formula, KState>>, ValuationSet>, List<Map<Tuple<Tuple<Formula, KState>, Tuple<Formula, KState>>, ValuationSet>>>(pair.left, newSucceedList));
        }
        Misc.verboseln(toString(step1));

        Misc.verboseln("2. Removing redundant reach sets by mapping each gen pair (F, I) |-> (F, { i | i in I and !\\exists j in I : j < i })\n");
        Set<Tuple<Map<Tuple<Tuple<Formula, KState>, Tuple<Formula, KState>>, ValuationSet>, List<Map<Tuple<Tuple<Formula, KState>, Tuple<Formula, KState>>, ValuationSet>>>> step2 = new HashSet<Tuple<Map<Tuple<Tuple<Formula, KState>, Tuple<Formula, KState>>, ValuationSet>, List<Map<Tuple<Tuple<Formula, KState>, Tuple<Formula, KState>>, ValuationSet>>>>();
        for (Tuple<Map<Tuple<Tuple<Formula, KState>, Tuple<Formula, KState>>, ValuationSet>, List<Map<Tuple<Tuple<Formula, KState>, Tuple<Formula, KState>>, ValuationSet>>> pair : step1) {
            List<Map<Tuple<Tuple<Formula, KState>, Tuple<Formula, KState>>, ValuationSet>> newSucceedList = new ArrayList<Map<Tuple<Tuple<Formula, KState>, Tuple<Formula, KState>>, ValuationSet>>();
            for (Map<Tuple<Tuple<Formula, KState>, Tuple<Formula, KState>>, ValuationSet> i : pair.right) {
                boolean redundant = false;
                for (Map<Tuple<Tuple<Formula, KState>, Tuple<Formula, KState>>, ValuationSet> j : pair.right) {
                    //if (i.entrySet().containsAll(j.entrySet()) && !i.equals(j))
                    boolean jSubsetOfi = true;
                    if (!i.equals(j)) {
                        for (Tuple<Tuple<Formula, KState>, Tuple<Formula, KState>> edgeKey : j.keySet()) {
                            if (!i.keySet().contains(edgeKey) || !i.get(edgeKey).containsAll(j.get(edgeKey))) {
                                jSubsetOfi = false;
                                break;
                            }
                        }
                        if (jSubsetOfi) {
                            redundant = true;
                            break;
                        }
                    }
                }
                if (redundant) {
                    Misc.verboseln("Redundant ReachSet" + i);
                } else if (newSucceedList.contains(i)) {
                    Misc.verboseln("Duplicate ReachSet" + i);
                } else {
                    newSucceedList.add(i);
                }
            }
            step2.add(new Tuple<Map<Tuple<Tuple<Formula, KState>, Tuple<Formula, KState>>, ValuationSet>, List<Map<Tuple<Tuple<Formula, KState>, Tuple<Formula, KState>>, ValuationSet>>>(pair.left, newSucceedList));
        }
        Misc.verboseln("\n" + toString(step2));

        Misc.verboseln("3. Removing pairs (F, {..., \\emptyset, ...} )\n");
        Set<Tuple<Map<Tuple<Tuple<Formula, KState>, Tuple<Formula, KState>>, ValuationSet>, List<Map<Tuple<Tuple<Formula, KState>, Tuple<Formula, KState>>, ValuationSet>>>> step3 = new HashSet<Tuple<Map<Tuple<Tuple<Formula, KState>, Tuple<Formula, KState>>, ValuationSet>, List<Map<Tuple<Tuple<Formula, KState>, Tuple<Formula, KState>>, ValuationSet>>>>();
        for (Tuple<Map<Tuple<Tuple<Formula, KState>, Tuple<Formula, KState>>, ValuationSet>, List<Map<Tuple<Tuple<Formula, KState>, Tuple<Formula, KState>>, ValuationSet>>> pair : step2) {
            // if (!pair.right.isEmpty())
            boolean add = true;
            for (Map<Tuple<Tuple<Formula, KState>, Tuple<Formula, KState>>, ValuationSet> c2 : pair.right) {
                if (c2.isEmpty()) {
                    add = false;
                    break;
                }
            }
            if (add) {
                step3.add(pair);
            }
        }
        Misc.verboseln(toString(step3));

        Misc.verboseln("4. Removing pairs (F, I) s.t. F = Q\n");
        Set<Tuple<Map<Tuple<Tuple<Formula, KState>, Tuple<Formula, KState>>, ValuationSet>, List<Map<Tuple<Tuple<Formula, KState>, Tuple<Formula, KState>>, ValuationSet>>>> step4 = new HashSet<Tuple<Map<Tuple<Tuple<Formula, KState>, Tuple<Formula, KState>>, ValuationSet>, List<Map<Tuple<Tuple<Formula, KState>, Tuple<Formula, KState>>, ValuationSet>>>>();
        for (Tuple<Map<Tuple<Tuple<Formula, KState>, Tuple<Formula, KState>>, ValuationSet>, List<Map<Tuple<Tuple<Formula, KState>, Tuple<Formula, KState>>, ValuationSet>>> pair : step3) {
            if (!pair.left.equals(kdfaA.edges)) {
                step4.add(pair);
            }
        }
        Misc.verboseln(toString(step4));

        Misc.verboseln("5. Removing pairs (F, I) s.t. there is (G, J) < (F, I)\n");
        Set<Tuple<Map<Tuple<Tuple<Formula, KState>, Tuple<Formula, KState>>, ValuationSet>, List<Map<Tuple<Tuple<Formula, KState>, Tuple<Formula, KState>>, ValuationSet>>>> step5 = new HashSet<Tuple<Map<Tuple<Tuple<Formula, KState>, Tuple<Formula, KState>>, ValuationSet>, List<Map<Tuple<Tuple<Formula, KState>, Tuple<Formula, KState>>, ValuationSet>>>>();
        for (Tuple<Map<Tuple<Tuple<Formula, KState>, Tuple<Formula, KState>>, ValuationSet>, List<Map<Tuple<Tuple<Formula, KState>, Tuple<Formula, KState>>, ValuationSet>>> p : step4) {
            boolean pIsRedundant = false;
            for (Tuple<Map<Tuple<Tuple<Formula, KState>, Tuple<Formula, KState>>, ValuationSet>, List<Map<Tuple<Tuple<Formula, KState>, Tuple<Formula, KState>>, ValuationSet>>> q : step4) {
                if (pairSubsumed(p, q) && !p.equals(q)) {
                    pIsRedundant = true;
                    break;
                }
            }
            if (!pIsRedundant) {
                step5.add(p);
            }
        }
        Misc.verboseln(toString(step5));

        condition = step5;
    }

    /**
     * True if p > q, i.e. p is more restrictive
     */
    private boolean pairSubsumed(Tuple<Map<Tuple<Tuple<Formula, KState>, Tuple<Formula, KState>>, ValuationSet>, List<Map<Tuple<Tuple<Formula, KState>, Tuple<Formula, KState>>, ValuationSet>>> p,
        Tuple<Map<Tuple<Tuple<Formula, KState>, Tuple<Formula, KState>>, ValuationSet>, List<Map<Tuple<Tuple<Formula, KState>, Tuple<Formula, KState>>, ValuationSet>>> q) {
        for (Map<Tuple<Tuple<Formula, KState>, Tuple<Formula, KState>>, ValuationSet> i : q.right) {
            //for (Set<Tuple<Formula, List<Set<Formula>>>> j : p.right) {
            boolean iCanBeMatched = false;
            for (Map<Tuple<Tuple<Formula, KState>, Tuple<Formula, KState>>, ValuationSet> j : p.right) {
                //for (Set<Tuple<Formula, List<Set<Formula>>>> i : q.right) {
                //if (i.entrySet().containsAll(j.entrySet())) {
                boolean jSubsetOfi = true;
                for (Tuple<Tuple<Formula, KState>, Tuple<Formula, KState>> edgeKey : j.keySet()) {
                    if (!i.keySet().contains(edgeKey) || !i.get(edgeKey).containsAll(j.get(edgeKey))) {
                        jSubsetOfi = false;
                        break;
                    }
                }
                if (jSubsetOfi) {
                    iCanBeMatched = true;
                    break;
                }
            }
            if (!iCanBeMatched) {
                return false;
            }
        }
        return p.left.entrySet().containsAll(q.left.entrySet());
    }

    public void setFormat(Integer form) {
        format = form;
    }

}
