package rabinizer.exec;

import java.util.ArrayList;
import java.lang.*;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.Stack;
import java.util.Map.Entry;

//import jdd.util.Array;
import rabinizer.formulas.Formula;

/**
 * 
 * @author Ruslan Ledesma-Garza
 *
 */

public class RabinAutomaton {
	
    public Set<Tuple<Tuple<Formula, List<Set<Formula>>>, List<Integer>>> states = null;
    public Tuple<Tuple<Formula, List<Set<Formula>>>, List<Integer>> initialState = null;

    public Map<Tuple<Tuple<Tuple<Formula, List<Set<Formula>>>, List<Integer>>, Tuple<Tuple<Formula, List<Set<Formula>>>, List<Integer>>>, ValuationSet> edges = null;
    public List<Tuple<Set<Tuple<Formula, List<Set<Formula>>>>, List<Set<Tuple<Formula, List<Set<Formula>>>>>>> condition = null;

    public Map<Formula, Formula> displayLabels = null;

    RabinAutomaton(DFA dfaA, GeneralizedRabinAcceptanceCondition grac) {
    	Misc.verboseln("Rabin Automaton");
    	
    	displayLabels = dfaA.displayLabels;
    	
    	// fix the order of the pairs and the order of the reach eventually
    	condition = new ArrayList<Tuple<Set<Tuple<Formula, List<Set<Formula>>>>, List<Set<Tuple<Formula, List<Set<Formula>>>>>>>();
    	//Repair condition!
    	/*for (Tuple<Set<Tuple<Formula, List<Set<Formula>>>>, Set<Set<Tuple<Formula, List<Set<Formula>>>>>> pair : grac.condition){
    		List<Set<Tuple<Formula, List<Set<Formula>>>>> reachEventually = new ArrayList<Set<Tuple<Formula, List<Set<Formula>>>>>(pair.right.size());
    		reachEventually.addAll(pair.right);
    		condition.add(new Tuple<Set<Tuple<Formula, List<Set<Formula>>>>, List<Set<Tuple<Formula, List<Set<Formula>>>>>>(pair.left, reachEventually));
    	}
    	*/
    	Misc.verboseln("Generalized Rabin Acceptance Condition w/ fixed order");
    	Misc.verboseln(""+condition);
    	
    	// initial state
    	Misc.verboseln("Number of copies of DFA: " + condition.size());
    	List<Integer> iSCoords = new ArrayList<Integer>(condition.size());
		for (int i = 0; i < condition.size(); i++) {
    		iSCoords.add(i, 0);
    	}
    	initialState = new Tuple<Tuple<Formula, List<Set<Formula>>>, List<Integer>>(dfaA.initialState, iSCoords);

        states = new HashSet<Tuple<Tuple<Formula, List<Set<Formula>>>, List<Integer>>>();
        states.add(initialState);
        edges = new HashMap<Tuple<Tuple<Tuple<Formula, List<Set<Formula>>>, List<Integer>>, Tuple<Tuple<Formula, List<Set<Formula>>>, List<Integer>>>, ValuationSet>();
        Stack<Tuple<Tuple<Formula, List<Set<Formula>>>, List<Integer>>> workstack = new Stack<Tuple<Tuple<Formula, List<Set<Formula>>>, List<Integer>>>();
        workstack.push(initialState);

        int loopCount = 0;
        while (!workstack.empty()) {
        
        	Misc.verboseln("loop count: " + loopCount++);
        	Misc.verboseln("workstack : " + workstack);
        	Misc.verboseln("workcount : " + workstack.size());
        	
        	Tuple<Tuple<Formula, List<Set<Formula>>>, List<Integer>> curr = workstack.pop();
            Misc.verboseln("Curr: " + curr);
            for (Entry<Tuple<Tuple<Formula, List<Set<Formula>>>, Tuple<Formula, List<Set<Formula>>>>, ValuationSet> dfaEdge : dfaA.edges.entrySet()) {
            	if (dfaEdge.getKey().left.equals(curr.left)) {
	                Misc.verboseln("\tDFA A edge: " + dfaEdge);
	                // successor state
	                Tuple<Tuple<Formula, List<Set<Formula>>>, List<Integer>> succ = step(curr, dfaEdge.getKey());
	                Misc.verboseln("\tSucc: " + succ);
	                // state exploration
	                if (!states.contains(succ)) {
	                	Misc.verboseln("\tNew state");
	                    states.add(succ);
	                    workstack.push(succ);
	                }
	                // edge
	                Tuple<Tuple<Tuple<Formula, List<Set<Formula>>>, List<Integer>>, Tuple<Tuple<Formula, List<Set<Formula>>>, List<Integer>>> currAndSucc = new Tuple<Tuple<Tuple<Formula, List<Set<Formula>>>, List<Integer>>, Tuple<Tuple<Formula, List<Set<Formula>>>, List<Integer>>>(curr, succ);
	                edges.put(currAndSucc, dfaEdge.getValue());
	                Misc.verboseln("\tLabel of edge: " + dfaEdge.getValue());
            	}
            }
        }
        Misc.verboseln("Initial state: " + initialState);
        Misc.verboseln("Edges:         " + edges);
    }
    
    
    private Tuple<Tuple<Formula, List<Set<Formula>>>, List<Integer>> step(Tuple<Tuple<Formula, List<Set<Formula>>>, List<Integer>> curr, Tuple<Tuple<Formula, List<Set<Formula>>>, Tuple<Formula, List<Set<Formula>>>> edge) {
    	List<Integer> coords = new ArrayList<Integer>(condition.size());
    	for (int i = 0; i < condition.size(); i++) {
    		int mi = curr.right.get(i);
    		Set<Tuple<Formula, List<Set<Formula>>>> reachEventuallyERF = condition.get(i).right.get(mi);
    		if (reachEventuallyERF.contains(curr.left)) {
    			coords.add(i, (mi + 1) % condition.get(i).right.size());
    		} else {
    			coords.add(i, mi);
    		}
    	}
    	return new Tuple<Tuple<Formula, List<Set<Formula>>>, List<Integer>>(edge.right, coords);
    }
    
    public String toDotty() {
        // The initial state in the set states may be different from the states in the set of edges. This can happen
        // because HashSet does not guarantee that two segment states that are equal have the same string representation.
        // To circumvent the problem, use the same string representation for rendering nodes and edges in the dotty file.
        Map<Tuple<Tuple<Formula, List<Set<Formula>>>, List<Integer>>, Tuple<Tuple<Formula, List<Set<Formula>>>, List<Integer>>> normalStates = new HashMap<Tuple<Tuple<Formula, List<Set<Formula>>>, List<Integer>>, Tuple<Tuple<Formula, List<Set<Formula>>>, List<Integer>>>();        
        String dot = "digraph \"RabinAutomaton " + initialState +"\" {\n";
        for (Tuple<Tuple<Formula, List<Set<Formula>>>, List<Integer>> s : states) {
        	normalStates.put(s,s);
        	Misc.verbose(",");
        	Formula display = displayLabels.get(s.left.left);
        	if (s == initialState) {
                dot += "node [shape=oval, label=\"" + display + " :: " + s.left.right + "\\n" + s.right + "\"]\"" + s + "\";\n";
            } else {
                dot += "node [shape=rectangle, label=\"" + display + " :: " + s.left.right + "\\n" + s.right + "\"]\"" + s + "\";\n";
            }
        }
        for (Entry<Tuple<Tuple<Tuple<Formula, List<Set<Formula>>>, List<Integer>>, Tuple<Tuple<Formula, List<Set<Formula>>>, List<Integer>>>, ValuationSet> edge : edges.entrySet()) {
        	Misc.verbose(".");
            dot += "\"" + normalStates.get(edge.getKey().left) + "\" -> \"" + normalStates.get(edge.getKey().right) + "\" [label=\"" + edge.getValue() + "\"];\n"; 
        }
        return dot + "}\n";
    }
}
