package rabinizer.formulas;

import java.util.*;

import net.sf.javabdd.BDD;
import rabinizer.exec.*;

/**
 * @author Andreas Gaiser & Ruslan Ledesma-Garza
 *
 */
public class BooleanConstant extends Formula {

	public boolean value;

	public BooleanConstant(boolean value) {
		this.value = value;
		cachedBdd = (value ? Misc.bddFactory.one() : Misc.bddFactory.zero());
		Misc.representativeOfBdd(cachedBdd, this);
	}
	
	public BDD bdd() {
		return cachedBdd;
	}

	@Override
	public int hashCode() {
		return value ? 0 : 1;
	}

	@Override 
	public boolean equals(Object o) {
		if (!(o instanceof BooleanConstant))
			return false;
		else {
			return ((BooleanConstant)o).value == value;
		}
	}
	
	@Override
	public Formula negated() {
		return new BooleanConstant(!value);
	}

    	@Override
        public String toReversePolishString() {
            return toString();
        }
    
	@Override
	public String toString() {
		if (strFormula == null) strFormula = (value ? "tt" : "ff");
		return strFormula;
	}

	@Override
	protected Tuple<Set<Formula>, Set<Formula>> recurrentFormulas(boolean accumulate) {
            return new Tuple<Set<Formula>, Set<Formula>>(new HashSet<Formula>(), new HashSet<Formula>());
	}

	@Override
	public Formula evaluateCurrentAssertions(Valuation valuation) {
		return this;
	}
	
	@Override
	public Formula removeConstantCurrentAssertions() {
		return this;
	}

	@Override
	public Formula removeX() {
		return this;
	}
	
	@Override
	public Formula unfold() {
		return this;
	}

	@Override
	public Formula unfoldNoG() {
		return this;
	}
	
	@Override
	public boolean isProgressFormula() {
		return false;
	}

    	@Override
	public boolean hasSubformula(Formula f) {
            return this.equals(f);
	}

        @Override
        public Formula removeXsFromCurrentBooleanAtoms() {
            return this;
        }

    	@Override
            public Set<Formula> argumentsAlways() {
            return new HashSet<Formula>();
	}

       @Override
            public Set<Formula> argumentsFinsideG(boolean acc) {
            return new HashSet<Formula>();
	}
       
   	public boolean untilOcurrs() {
   		return false;
   	}
   	
   	public Formula toNNF() {
   		return this;
   	}

}
